function v=TPS_val(Xi,lambda,z,epsilon,m,m0,q)
%v=TPS_val(Xi,lambda,z,epsilon,m,m0,q) returns v=s(z), where s is the thin-plat rbf defined by
%s(z)=sum_j lambda(j) phi(z-Xi(j)) and phi(z)=|z|^2 log|z|.  It is assumed that Xi and z are
%complex row vectors and lambda is a (real) row vector. The optional input epsilon > 0, which
%defaults to 0.001, governs the accuracy of the returned row vector v, 
%whereby |s(z(j)) - v(j)| <= epsilon.  The optional inputs m and m0, which default to 12 and 10, 
%determine the maximal oscillatory degree of the moments used in the outer and inner summaries. 
%It is assumed that 3 <= m0 <= 25 and m0 <= m <= 40.  The optional input q >= 2, 
%which defaults to 27, is the minimum number of data sites in a cluster allowed to have children 
%(ie any cluster having less than q data sites will definitely not have children).

%______________________________________________________________________________________________ 
%                            Check that the inputs are acceptable

if nargin<3
  error("TPS_val: I require at least three inputs.");
end

if rows(Xi) != 1
  error("TPS_val: Xi should be a complex row vector.");
end
n=length(Xi);

if rows(lambda) != 1 || isreal(lambda)==0 || length(lambda)!=n
  error("TPS_val: lam should be a real row vector with the same length as Xi.");
end

if rows(z) != 1
  error("TPS_val: z should be a complex row vector.");
end

if nargin>3
  if isscalar(epsilon)==0 || epsilon < 0
     error("TPS_val: I require epsilon >= 0");
  end
else
  epsilon=0.001;
end

if nargin>4
  if isscalar(m)==0 || m!=round(m) || m<3 || m>40
    error("TPS_val: I requre m to be an integer with 3 <= m <= 40");
  end
else
  m=12;
end

if nargin>5
  if isscalar(m0)==0 || m0!=round(m0) || m0<3 || m0>m || m0>25
    error("TPS_val: I requre m0 to be an integer with 3 <= m0 <= min(m,25)");
  end
else
  m0=10;
end

if nargin>6
  if isscalar(q)==0 || q!=round(q) || q<2
    error("TPS_val: I requre q to be an integer >= 2");
  end
else
  q=27;
end

%________________________________________________________________________________________
% We load the auxiliary file TPS_val.dat which contains the matrices e_extrema,
% e_poly_coeff, a_coeff, b_coeff and crop them according to the choices of m0 and m

load TPS_val.dat;
e_extrema=e_extrema(:,m0); e_poly_coeff=e_poly_coeff(:,m0);
a_coeff=a_coeff(:,1:m0+1); b_coeff=b_coeff(:,1:m0+1);

%_______________________________________________________________________________
% Identify the square Omega_1 (a smallest square containing Xi)
% by determining its center c and radius r1
% (In the paper, r1 is called r0)

M1=max(real(Xi)); m1=min(real(Xi)); 
M2=max(imag(Xi)); m2=min(imag(Xi));
s=max(M1-m1,M2-m2); M1=m1+s; M2=m2+s;
c=m1+s/2 + I*(m2+s/2); r1=s/sqrt(2);

%_______________________________________________________________________________
%     Initialize N (number of pages in catalogue), CAT (integer fields of catalogue)
%                CENTER (cluster centres), H_mom (harmonic moments), 
%                biH_mom (biharmonic moments), and determine l_max (maximum allowed level).
%     Remark:  There is an implementation challenge caused by the fact that we do not know
%              apriori the final number of pages in the catalogue.  The approach taken here
%              is to let N_reserved denote the currently allocated (in memory) number of pages, 
%              while N denotes the current number of pages.  Whenever N reaches N_reserved+1, 
%              we extend the allocation to N_reserved=round(N_reserved*N_increase_factor).

N_reserved=20; N_increase_factor=1.2; 
CAT=zeros(8,N_reserved); CENTER=zeros(1,N_reserved); 
H_mom=zeros(m+2,N_reserved); biH_mom=H_mom;
norm_lambda=sum(abs(lambda));
l_max=ceil( log(r1^2 * e_extrema(2) * norm_lambda / epsilon) / log(4) );

%____________________________________________________________________________________________
%      Construction of the pages of the catalogue along with the moments of the childless pages.
%      Note 1: In addition to the variables just above, we also have the following in hand:
%           n (length of Xi), c (center of initial bounding square), r1 (radius of initial 
%           bounding square).  
%      Note 2: In the paper, the construction was described using a recursive function create_page.
%           In practice (particularly in Octave), it's better to use a stack rather than recursion,
%           so that's what's done here.
%           Each column of the stack holds the following details pertaining to a cluster
%           (who has yet to be registered in the catalogue): [n1 ; n2 ; l ; real_c ; imag_c ; p ; K]
%           where n1 and n2 are the first and last indices of the cluster, l is the level,
%           real_c + imag_c * I is the center of the bounding square, p is the page number of
%           the parent, and K in {I, II, III, IV} is the answer to "which child am I?"

stack=zeros(7,2);
sp=1; N=0;
stack(:,sp)=[1;length(lambda);0;real(c);imag(c);0;0];

% Here are a couple of handy vectors:
r=r1*(2.^-(0:l_max))'; r_alt=r/(2*sqrt(2));  % CAREFUL: these are supposed to be column vectors
                                             % (ie verify that 0:l_max is a row vector)
while(sp>0)
n1=stack(1,sp); n2=stack(2,sp); l=stack(3,sp); real_c=stack(4,sp); imag_c=stack(5,sp);
N += 1;

if N>N_reserved
  N_reserved=round(N_reserved*N_increase_factor); 
  CAT(8,N_reserved)=0; CENTER(N_reserved)=0; H_mom(m+2,N_reserved)=0;  % this will cause Octave
  biH_mom(m+2,N_reserved)=0;                                           % to extend these objects
end

if N>1
  CAT(stack(7,sp),stack(6,sp))=N;
end
CENTER(N)=real_c+I*imag_c;
sp-=1;

CAT(1:8,N)=[0;0;0;0;0;n1;n2;l];

if n2-n1 > q && l<l_max % We're having children
 I1=real(Xi(n1:n2))>real_c; I2=imag(Xi(n1:n2))>imag_c; 
 J1=find(I1.*I2); J2=find((1-I1).*I2);
 J3=find((1-I1).*(1-I2)); J4=find((1-I2).*I1);
 Xi(n1:n2)=Xi(n1-1+[J1,J2,J3,J4]); 
 lambda(n1:n2)=lambda(n1-1+[J1,J2,J3,J4]); 
 k=n2+1; s=r_alt(l+1); % saved value of s=r1*2^(-l-1.5);

 n_ch=length(J4);
 if n_ch>0 
   sp+=1; stack(:,sp)=[k-n_ch;k-1;l+1;real_c+s;imag_c-s;N;4]; 
   k-=n_ch; CAT(5,N)+=1;;
 end

 n_ch=length(J3);
 if n_ch>0 
   sp+=1; stack(:,sp)=[k-n_ch;k-1;l+1;real_c-s;imag_c-s;N;3]; 
   k-=n_ch; CAT(5,N)+=1;
 end

 n_ch=length(J2);
 if n_ch>0 
   sp+=1; stack(:,sp)=[k-n_ch;k-1;l+1;real_c-s;imag_c+s;N;2]; 
   k-=n_ch; CAT(5,N)+=1;
 end

 n_ch=length(J1);
 if n_ch>0 
   sp+=1; stack(:,sp)=[k-n_ch;k-1;l+1;real_c+s;imag_c+s;N;1]; 
   k-=n_ch; CAT(5,N)+=1;
 end
 
else
  [H_mom(:,N),biH_mom(:,N)]=direct_moments_fast((Xi(n1:n2)-CENTER(N))/r(l+1),lambda(n1:n2),m);
end

end %while(sp>0)

if N<N_reserved
  CAT=CAT(:,1:N); CENTER=CENTER(:,1:N); H_mom=H_mom(:,1:N); biH_mom=biH_mom(:,1:N);
end


%_______________________________________________________________________________
%          Construct the lower triangular matrices L1, L2, L3, L4, and then
%          the moments of clusters having children

z1=exp(I*(pi/4)*(0:7)); z2=exp(I*(3*pi/4)*(0:7));
z3=exp(I*(-3*pi/4)*(0:7)); z4=exp(I*(-pi/4)*(0:7));
L1=zeros(m+2,m+2); L2=L1; L3=L1; L4=L1;
for i=0:m+1
 temp1=2^-i;
 for j=0:i
   temp2=nchoosek(i,j)*temp1;
   L1(i+1,j+1)=temp2*z1(1+mod(i-j,8));
   L2(i+1,j+1)=temp2*z2(1+mod(i-j,8));
   L3(i+1,j+1)=temp2*z3(1+mod(i-j,8));
   L4(i+1,j+1)=temp2*z4(1+mod(i-j,8));
 end
end
clear z1 z2 z3 z4;

% Now we construct the moments

for i=N:-1:1
  if CAT(5,i)>0
    if (j=CAT(4,i))>0
      temp=L4*H_mom(:,j); H_mom(:,i)+=temp;
      biH_mom(:,i)+=conj(L4(2,1))*temp+0.5*(L4*biH_mom(:,j));      
    end
    if (j=CAT(3,i))>0
      temp=L3*H_mom(:,j); H_mom(:,i)+=temp;
      biH_mom(:,i)+=conj(L3(2,1))*temp+0.5*(L3*biH_mom(:,j));      
    end
    if (j=CAT(2,i))>0
      temp=L2*H_mom(:,j); H_mom(:,i)+=temp;
      biH_mom(:,i)+=conj(L2(2,1))*temp+0.5*(L2*biH_mom(:,j));      
    end
    if (j=CAT(1,i))>0
      temp=L1*H_mom(:,j); H_mom(:,i)+=temp;
      biH_mom(:,i)+=conj(L1(2,1))*temp+0.5*(L1*biH_mom(:,j));      
    end
  end
end

%   The moments have been constructed in the order mentioned in the paper (this allows
%   the moments of a parent to be obtained from those of its children by matrix multiplication).
%   We now extract the desired moments:
H_mom=H_mom(1:m+1,:);  biH_mom=biH_mom(2:m+2,:);

%   It's possible that the maximum level in the obtained catalogue is less than l_max, so
%   if that happens, we will adjust l_max accordingly.
max_level=max(CAT(8,:));
if max_level<l_max
  disp(["theoretical value of l_max is ",num2str(l_max),", but maximum level in catalogue is ",num2str(max_level)]);
  l_max=max_level; r=r(1:l_max+1);
end

%______________________________________________________________________________________________
%   We now form the appendix of the catalogue, which is a 3 x (l_max + 1) matrix, where
%   column l+1 contains [t_l^2 ; r_l^2 ; T_l^2].  Here l denotes the level of a page,
%   r_l = r1/2^l is the radius of the bounding square, t_l is the inner threshold radius and
%   T_l is the outer threshold radius.

%   The inner threshold radius t_l is defined in terms of a monotone decreasing function e_m0
%   which can be viewed as a range limited polynomial.  The polynomial, whose coefficients
%   are stored in e_poly_coeff, is decreasing on [0,1], however the range of e_m0 is restricted
%   to the interval [e_min,e_max] (stored in e_extrema as [e_min ; e_max]).  Mathematically, we
%   define t_l = r_l * min ( {1} U {s in [0,1) : e_m0(s) <= rhs} ).  In practice, we do:
%   Put rhs = epsilon/(norm_lambda * r_l^2).
%     If rhs >= e_max, then t_l=0    (ie inner summaries are always used).
%     If rhs <= e_min, then t_l=r_l  (ie inner summaries are never used).
%     Otherwise, let s in (0,1) be the solution of e_m0(s) = rhs, and put t_l = r_l * s.
%                                    (ie inner summaries are used when t_l <= |z-c| < r_l)
%
%   Incidentally, the equation e_m0(s) = rhs above is solved using the secant method with
%   initial values s_0=0, s_1=1.

%   The outer threshold radius T_l is defined in terms of the strictly decreasing function 
%                   E_m(t) = t^(1-m)/( m(m+1) ) + t^(-m)/( (m+1)(m+2) )
%   Mathematically, we define T_l = r_l * min {s >= 1 : E_m(s) <= rhs}.  In practice, we do:
%   Put rhs = epsilon/(norm_lambda * r_l^2).
%     If rhs >= E_m(1), then T_l=r_l    (ie outer summaries are always used).
%   Otherwise, let s > 1 be the solution of E_m(s) = rhs (found by Newton's method with s0=1),
%   and put T_l = r_l * s.          (ie outer summaries are used when T_l <= |z-c| )

Appendix = zeros(l_max+1,3); Appendix(:,2)=r.^2;

for l=0:l_max
  rhs=epsilon/(norm_lambda * Appendix(l+1,2));
  if e_extrema(1)>=rhs
    Appendix (l+1,1)=Appendix(l+1,2); % ie t_l = r_l
  elseif e_extrema(2)<=rhs
    Appendix (l+1,1)=0; % ie t_l = 0
  else
    % here comes the secant method:
    tol=1e-7;
    s0=0; s1=1; i=0;
    f0=polyval(e_poly_coeff,s0)-rhs;
    while abs(s1-s0)>=tol && i<100
      f1=polyval(e_poly_coeff,s1)-rhs;
      s=s1-f1*(s1-s0)/(f1-f0);
      s0=s1; f0=f1; 
      s1=s; i+=1;
    end
    if s<0 || s>1
       disp("TPS_val: Secant method isn't able to solve e_n(s) = rhs");
       keyboard;
    end
    Appendix(l+1,1)=Appendix(l+1,2)*s^2; % ie t_l = r_l * s
  end

  if 2/(m*(m+2))<=rhs
    Appendix (l+1,3)=Appendix(l+1,2); % ie T_l = r_l
  else
    % here comes Newton's method:
    s=0; s1=1; tol=1e-10; c=m*(m+1)*(m+2)*rhs;
    while(abs(s1-s)>tol)
      s=s1;
      s1=s+( (m+2)*s + m -c*s^m)/( (m+2)*(m-1) + m*m/s );
    end
    if s1<1
       disp("TPS_val: Newton's method is unable to solve Cal_E(s) = rhs");
       keyboard;
    end
    Appendix(l+1,3)=Appendix(l+1,2)*s1^2; % ie T_l = r_l * s
  end
end

%____________________________________________________________________________________________
%     Ad-hoc optimizations:
% It is possible to further simplify the evaluation phase by 
% 1. appending two columns to Appendix, the first contains r_ell and the second r_ell^2 log(r_ell)

Appendix=[Appendix,r,Appendix(:,2).*log(r)];

% 2. pre-divide H_mom and biH_mom to simplify the outer summary computations
%    and pre-multiply a_coeff and b_coeff so that inner summaries are not affected.

% For k=2,3,...,m
% we divide alpha_k by k(k-1) and multiply a_k(s) by k(k-1)
% And we divide beta_k by k(k+1) and multiply b_k(s) by k(k+1)
for k=2:m0
  temp=k*(k-1); temp_inv=1/temp;
  H_mom(k+1,:)*=temp_inv; a_coeff(:,k+1)*=temp;
  temp=k*(k+1); temp_inv=1/temp;
  biH_mom(k+1,:)*=temp_inv; b_coeff(:,k+1)*=temp;
end

for k=m0+1:m
  H_mom(k+1,:)*=1/(k*(k-1));
  biH_mom(k+1,:)*=1/(k*(k+1));
end

v=evaluate_fast(Xi,lambda,z,m,m0,CAT,Appendix,CENTER,H_mom,biH_mom,a_coeff,b_coeff);

